
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "gda-connection.h" */

GType gda_connection_error_get_type (void);
#define GDA_TYPE_CONNECTION_ERROR (gda_connection_error_get_type())

GType gda_connection_status_get_type (void);
#define GDA_TYPE_CONNECTION_STATUS (gda_connection_status_get_type())

GType gda_connection_options_get_type (void);
#define GDA_TYPE_CONNECTION_OPTIONS (gda_connection_options_get_type())

GType gda_connection_feature_get_type (void);
#define GDA_TYPE_CONNECTION_FEATURE (gda_connection_feature_get_type())

GType gda_connection_meta_type_get_type (void);
#define GDA_TYPE_CONNECTION_META_TYPE (gda_connection_meta_type_get_type())

/* enumerations from "gda-enums.h" */

GType gda_transaction_isolation_get_type (void);
#define GDA_TYPE_TRANSACTION_ISOLATION (gda_transaction_isolation_get_type())

GType gda_value_attribute_get_type (void);
#define GDA_TYPE_VALUE_ATTRIBUTE (gda_value_attribute_get_type())

GType gda_sql_identifier_style_get_type (void);
#define GDA_TYPE_SQL_IDENTIFIER_STYLE (gda_sql_identifier_style_get_type())

G_END_DECLS

/* Generated data ends here */

