/* gda-db-fkey-private.h
 *
 * Copyright (C) 2020 Pavlo Solntsev <p.sun.fun@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#ifndef GDA_DB_FKEY_PRIVATE_H
#define GDA_DB_FKEY_PRIVATE_H

#include "gda-db-fkey.h"
#include "gda-meta-struct.h"

GdaDbFkey       *gda_db_fkey_new_from_meta    (GdaMetaTableForeignKey *metafkey);

#endif /* end of include guard: GDA_DB_FKEY_PRIVATE_H_RX2KUYCG */



