/*
 * This file is generated by the gen_def program (see the gen_def.c file 
 * for some explanations)
 * DO NOT EDIT MANUALLY
 */


#define L_ILLEGAL 		 0
#define L_SQLCOMMENT 		 1
#define L_SPACE 		 2
#define L_ID 		 3
#define L_ABORT 		 4
#define L_AFTER 		 5
#define L_ANALYZE 		 6
#define L_ASC 		 7
#define L_ATTACH 		 8
#define L_BEFORE 		 9
#define L_BEGIN 		 10
#define L_CASCADE 		 11
#define L_CAST 		 12
#define L_CONFLICT 		 13
#define L_DATABASE 		 14
#define L_DEFERRED 		 15
#define L_DESC 		 16
#define L_DETACH 		 17
#define L_EACH 		 18
#define L_END 		 19
#define L_EXCLUSIVE 		 20
#define L_EXPLAIN 		 21
#define L_FAIL 		 22
#define L_FOR 		 23
#define L_IGNORE 		 24
#define L_IMMEDIATE 		 25
#define L_INITIALLY 		 26
#define L_INSTEAD 		 27
#define L_LIKE 		 28
#define L_ILIKE 		 29
#define L_MATCH 		 30
#define L_PLAN 		 31
#define L_QUERY 		 32
#define L_KEY 		 33
#define L_OF 		 34
#define L_OFFSET 		 35
#define L_PRAGMA 		 36
#define L_RAISE 		 37
#define L_REPLACE 		 38
#define L_RESTRICT 		 39
#define L_ROW 		 40
#define L_TEMP 		 41
#define L_TRIGGER 		 42
#define L_VACUUM 		 43
#define L_VIEW 		 44
#define L_VIRTUAL 		 45
#define L_REINDEX 		 46
#define L_RENAME 		 47
#define L_CTIME_KW 		 48
#define L_IF 		 49
#define L_DELIMITER 		 50
#define L_COMMIT 		 51
#define L_ROLLBACK 		 52
#define L_ISOLATION 		 53
#define L_LEVEL 		 54
#define L_SERIALIZABLE 		 55
#define L_READ 		 56
#define L_COMMITTED 		 57
#define L_UNCOMMITTED 		 58
#define L_REPEATABLE 		 59
#define L_WRITE 		 60
#define L_ONLY 		 61
#define L_SAVEPOINT 		 62
#define L_RELEASE 		 63
#define L_COMMENT 		 64
#define L_FORCE 		 65
#define L_WAIT 		 66
#define L_NOWAIT 		 67
#define L_BATCH 		 68
#define L_TEXTUAL 		 69
#define L_STRING 		 70
#define L_OR 		 71
#define L_AND 		 72
#define L_NOT 		 73
#define L_IS 		 74
#define L_NOTLIKE 		 75
#define L_NOTILIKE 		 76
#define L_IN 		 77
#define L_ISNULL 		 78
#define L_NOTNULL 		 79
#define L_DIFF 		 80
#define L_EQ 		 81
#define L_BETWEEN 		 82
#define L_GT 		 83
#define L_LEQ 		 84
#define L_LT 		 85
#define L_GEQ 		 86
#define L_REGEXP 		 87
#define L_REGEXP_CI 		 88
#define L_NOT_REGEXP 		 89
#define L_NOT_REGEXP_CI 		 90
#define L_SIMILAR 		 91
#define L_ESCAPE 		 92
#define L_BITAND 		 93
#define L_BITOR 		 94
#define L_LSHIFT 		 95
#define L_RSHIFT 		 96
#define L_PLUS 		 97
#define L_MINUS 		 98
#define L_STAR 		 99
#define L_SLASH 		 100
#define L_REM 		 101
#define L_CONCAT 		 102
#define L_COLLATE 		 103
#define L_UMINUS 		 104
#define L_UPLUS 		 105
#define L_BITNOT 		 106
#define L_LP 		 107
#define L_RP 		 108
#define L_JOIN 		 109
#define L_INNER 		 110
#define L_NATURAL 		 111
#define L_LEFT 		 112
#define L_RIGHT 		 113
#define L_FULL 		 114
#define L_CROSS 		 115
#define L_UNION 		 116
#define L_EXCEPT 		 117
#define L_INTERSECT 		 118
#define L_PGCAST 		 119
#define L_SEMI 		 120
#define L_END_OF_FILE 		 121
#define L_TRANSACTION 		 122
#define L_COMMA 		 123
#define L_INTEGER 		 124
#define L_TO 		 125
#define L_INSERT 		 126
#define L_INTO 		 127
#define L_VALUES 		 128
#define L_DELETE 		 129
#define L_FROM 		 130
#define L_WHERE 		 131
#define L_UPDATE 		 132
#define L_SET 		 133
#define L_ALL 		 134
#define L_SELECT 		 135
#define L_LIMIT 		 136
#define L_ORDER 		 137
#define L_BY 		 138
#define L_HAVING 		 139
#define L_GROUP 		 140
#define L_USING 		 141
#define L_ON 		 142
#define L_OUTER 		 143
#define L_DOT 		 144
#define L_AS 		 145
#define L_DISTINCT 		 146
#define L_CASE 		 147
#define L_WHEN 		 148
#define L_THEN 		 149
#define L_ELSE 		 150
#define L_NULL 		 151
#define L_FLOAT 		 152
#define L_UNSPECVAL 		 153
#define L_LSBRACKET 		 154
#define L_RSBRACKET 		 155
#define L_SIMPLEPARAM 		 156
#define L_PNAME 		 157
#define L_PDESCR 		 158
#define L_PTYPE 		 159
#define L_PNULLOK 		 160
#define L_RAWSTRING 		 161
#define L_LOOP 		 162
#define L_ENDLOOP 		 163
#define L_DECLARE 		 164
#define L_CREATE 		 165
#define L_BLOB 		 166
gint parser_tokens[] = {
118,119,118,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,120,121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,118,118,118,118,118,118};
gint delim_tokens[] = {
7,8,11,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,13,12,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9,10,1,1,14,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,15,16,17,18,19,20,21,22,23,1,2,3,4,5,6};
